@@ -2,13 +2,18 @@
 #include "config.h"
 #endif
 
-#include <glade/glade.h>   
 #include <libintl.h>
 #include <sys/wait.h>
+#include <glob.h>
+
+#include <glib.h>
+#include <glib/gstdio.h>
 
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <stdlib.h>
+#include <dbus/dbus-glib.h>
 
 #include "update-notifier.h"
 #include "update.h"
@@ -18,26 +23,43 @@
 // command, description, desktopfile, needs_gksu
 const char* actions[][4] = {
    { "/usr/bin/update-manager", N_("Show updates"), 
-     "/usr/share/applications/update-manager.desktop", FALSE },
+     "/usr/share/applications/update-manager.desktop", 
+     GINT_TO_POINTER(FALSE) },
 
    { "/usr/sbin/synaptic --dist-upgrade-mode --non-interactive --hide-main-window -o Synaptic::AskRelated=true",
-     N_("Install all updates"), "/usr/share/applications/synaptic.desktop", TRUE
+     N_("Install all updates"), "/usr/share/applications/synaptic.desktop", 
+     GINT_TO_POINTER(TRUE)
    },
    { "/usr/sbin/synaptic --update-at-startup --non-interactive --hide-main-window", 
-     N_("Check for updates"), "/usr/share/applications/synaptic.desktop", TRUE },
+     N_("Check for updates"), "/usr/share/applications/synaptic.desktop", 
+     GINT_TO_POINTER(TRUE) },
    { "/usr/sbin/synaptic", N_("Start package manager"), 
-     "/usr/share/applications/synaptic.desktop", TRUE},
+     "/usr/share/applications/synaptic.desktop", 
+     GINT_TO_POINTER(TRUE)},
    { NULL, NULL, NULL }
 };
 
+enum { 
+   NOTIFICATION_DEFAULT, 
+   NOTIFICATION_IGNORE, 
+   NOTIFICATION_SHOW_UPDATES 
+};
 
-enum { NOTIFICATION_DEFAULT, NOTIFICATION_IGNORE, NOTIFICATION_SHOW_UPDATES };
+static inline void
+g_debug_update(const char *msg, ...)
+{
+   va_list va;
+   va_start(va, msg);
+   g_logv("update",G_LOG_LEVEL_DEBUG, msg, va);
+   va_end(va);
+}
 
 static gboolean
 activate_cb (GtkWidget *widget, 
 	     TrayApplet *ta)
 {
-   int index = 0;
+   UpdateTrayAppletPrivate *priv = (UpdateTrayAppletPrivate*)ta->user_data;
+   int index = priv->apt_is_running ? 3 : 0;
    
    // on double click we get two activate signals, so slow down things
    // here a bit to avoid double starting
@@ -120,7 +142,6 @@ void
 update_trayicon_create_menu(TrayApplet *ta)
 {
 	GtkWidget *menuitem;
-	GtkAccelGroup *accelgroup;
 	int i;
 
 	ta->menu = gtk_menu_new ();
@@ -147,7 +168,7 @@ update_trayicon_create_menu(TrayApplet *ta)
 			 G_CALLBACK(cb_toggled_show_notifications), ta);
 	
 
-	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_PREFERENCES, accelgroup);
+	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_PREFERENCES, NULL);
 	gtk_menu_shell_append (GTK_MENU_SHELL (ta->menu), menuitem);
 	g_signal_connect(G_OBJECT(menuitem), "activate", 
 			 G_CALLBACK(cb_preferences), (void*)ta);
@@ -159,11 +180,18 @@ update_trayicon_create_menu(TrayApplet *ta)
 void 
 update_apt_is_running(TrayApplet *ta, gboolean is_running)
 {
+   g_debug_update("update_apt_is_running: %i\n",is_running);
    if(ta == NULL)
       return;
 
+   // update internal status
    UpdateTrayAppletPrivate *priv = (UpdateTrayAppletPrivate*)ta->user_data;
-   //g_print("update_apt_is_running: %i\n",is_running);
+   priv->apt_is_running = is_running;
+
+   // if the user wants auto-launch mode, do not show the icon
+   // the auto launch stuff will do its magic later
+   if(gconf_client_get_bool(priv->gconf, GCONF_KEY_AUTO_LAUNCH, NULL))
+      return;
 
    if(is_running) {
 
@@ -195,9 +223,6 @@ update_apt_is_running(TrayApplet *ta, gboolean is_running)
    } else {
       gtk_status_icon_set_from_icon_name(ta->tray_icon, ta->name);
    }
-
-   priv->apt_is_running = is_running;
-
 }
 
 // actually show the notification 
@@ -224,14 +249,23 @@ show_notification(gpointer user_data)
       return TRUE;
 
    // now show a notification handle 
+   gchar *updates;
+   updates = g_strdup_printf(ngettext("There is %i update available. "
+				      "Click on the notification "
+				      "icon to show the "
+				      "available update.",
+				      "There are %i updates available. "
+				      "Click on the notification "
+				      "icon to show the "
+				      "available updates.",
+				      priv->num_upgrades), 
+			     priv->num_upgrades);
    NotifyNotification *n = notify_notification_new_with_status_icon(
 			      _("Software updates available"),
-			      _("Click on the notification"
-				" icon to show the "
-				"available updates.\n"),
+			      updates,
 			      NULL, 
 			      ta->tray_icon);
- 
+
    GdkPixbuf* pix= gtk_icon_theme_load_icon(gtk_icon_theme_get_default(), 
 					    GTK_STOCK_DIALOG_INFO, 48,0,NULL);
    notify_notification_set_icon_from_pixbuf (n, pix);
@@ -242,6 +276,7 @@ show_notification(gpointer user_data)
    priv->active_notification = n;
 
    // remove this from the timeout now
+   g_free(updates);
    return FALSE;
 }
 
@@ -260,8 +295,13 @@ outdated_nag(TrayApplet *ta)
    if ((stat("/var/lib/apt/periodic/update-success-stamp", &buf) == 0) &&
        (time(NULL) - buf.st_mtime > OUTDATED_NAG_AGE) ) {
       gtk_status_icon_set_visible (ta->tray_icon, TRUE);
-      ta->name = "gtk-dialog-warning";
-      gtk_status_icon_set_from_icon_name(ta->tray_icon, ta->name);
+      ta->name = "gtk-dialog-warning-panel";
+      GdkPixbuf *src = gtk_icon_theme_load_icon(gtk_icon_theme_get_default(),
+						ta->name, 48, 
+						GTK_ICON_LOOKUP_GENERIC_FALLBACK, 
+						NULL);
+      gtk_status_icon_set_from_pixbuf(ta->tray_icon, src);
+      g_object_unref(src);
       gtk_status_icon_set_tooltip(ta->tray_icon,
 				  _("The update information is outdated. "
 				    "This may be caused by network "
@@ -277,17 +317,234 @@ outdated_nag(TrayApplet *ta)
    return FALSE;
 }
 
+// use ubuntu-system-service (if available) to check
+// if the dpkg lock is taken currently or not
+//
+// if uncertain, return FALSE 
+static gboolean
+dpkg_lock_is_taken ()
+{
+   DBusGConnection *connection;
+   GError *error;
+   DBusGProxy *proxy;
+   gboolean locked = FALSE;
+  
+   error = NULL;
+   connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
+   if (connection == NULL) {
+      g_debug_update ("Failed to open connection to bus: %s\n", error->message);
+      g_error_free (error);
+      return FALSE;
+   }
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+                                     "com.ubuntu.SystemService",
+				     "/",
+				     "com.ubuntu.SystemService");
+  error = NULL;
+  if (!dbus_g_proxy_call (proxy, "is_package_system_locked", &error, 
+			  G_TYPE_INVALID,
+                          G_TYPE_BOOLEAN, &locked, G_TYPE_INVALID)) {
+     g_debug_update ("error during dbus call: %s\n", error->message);
+     g_error_free (error);
+     g_object_unref (proxy);
+     return FALSE;
+  }
+  g_object_unref (proxy);
+
+  g_debug_update ("is_package_system_locked: %i", locked);
+  return locked;
+}
+
+#if 0
+// ask devicekit.Power if we run on battery
+static gboolean
+running_on_battery ()
+{
+   DBusGConnection *connection;
+   GError *error;
+   DBusGProxy *proxy;
+   gboolean on_battery = FALSE;
+  
+   error = NULL;
+   connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
+   if (connection == NULL) {
+      g_debug_update ("Failed to open connection to bus: %s\n", error->message);
+      g_error_free (error);
+      return FALSE;
+   }
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+				     "org.freedesktop.DeviceKit.Power", 
+				     "/org/freedesktop/DeviceKit/Power",
+				     "org.freedesktop.DBus.Properties");
+  error = NULL;
+  if (!dbus_g_proxy_call (proxy, "Get", &error, 
+			  G_TYPE_STRING, "org.freedesktop.DeviceKit.Power", 
+			  G_TYPE_STRING, "on_battery",
+			  G_TYPE_INVALID,
+                          G_TYPE_BOOLEAN, &on_battery, 
+			  G_TYPE_INVALID)) {
+     g_debug_update ("failed Get dbus call: %s\n", 
+		     error?error->message:"no error given");
+     if(error)
+	g_error_free (error);
+     return FALSE;
+  }
+  g_debug_update ("on_battery: %i", on_battery);
+  return on_battery;
+}
+#endif
+
+// check if the security auto launch interval is over
+// and if the user is not using auto install of security
+// updates
+static gboolean
+auto_launch_security_now(UpdateTrayAppletPrivate *priv, 
+			 time_t now, 
+			 time_t last_launch)
+{
+   char *cmd[3];
+   int ret;
+   GError *error = NULL;
+
+   // no security updates
+   if (priv->num_security == 0)
+      return FALSE;
+	 
+   // security updates, but already launched recently
+   if ((last_launch + AUTOLAUNCH_MINIMAL_SECURITY_INTERVAL) > now) {
+      g_debug_update("security updates, but update-manager was launched "
+		     "within the AUTOLAUNCH_MINIMAL_SECURITY_INTERVAL\n");
+      return FALSE;
+   }
+
+   // run apt_check.py in "check if we install updates unattended" mode
+   // a exit status > 0 indicates that its unattended-upgrades is used
+   if ( g_file_test ("/usr/bin/unattended-upgrades", 
+		     G_FILE_TEST_IS_EXECUTABLE) ) {
+      cmd[0] = UPGRADE_CHECKER;
+      cmd[1] = "--security-updates-unattended";
+      cmd[2] = NULL;
+      if (g_spawn_sync("/", cmd, NULL, 0, NULL, NULL, 
+		       NULL, NULL, &ret, &error)) {
+	 g_debug("--security-updates-unattended: %i\n", WEXITSTATUS(ret));
+	 if( WEXITSTATUS(ret) > 0 )
+	    return FALSE;
+      } else if (error) {
+	 g_print("error: %s\n", error->message);
+      }
+   }
+
+   g_debug_update("security updates, auto-launching");
+   return TRUE;
+}
+
+// check the logs of dpkg and apt for timestamps, the idea
+// is to not auto launch if dpkg/apt were run manually by the
+// user
+static time_t
+newest_log_file_timestamp()
+{
+   struct stat buf;
+   int i, j;
+   time_t newest_log_stamp = 0;
+
+   char *log_glob[] = { "/var/log/dpkg.log*",
+		    "/var/log/apt/term.log*",
+		    NULL };
+
+   for (i=0; log_glob[i] != NULL; i++) 
+   {
+      glob_t pglob;
+      if(glob(log_glob[i], 0, NULL, &pglob) != 0) {
+	 g_warning("error from glob %s\n", log_glob[i]);
+	 continue;
+      }
+      for(j=0;j < pglob.gl_pathc; j++) {
+
+	 const char *log = pglob.gl_pathv[j];
+	 if(g_stat(log, &buf) <0) {
+	    g_warning("can't stat %s\n", log);
+	    continue;
+	 }
+	 if(buf.st_size == 0) {
+	    g_warning("log file empty (logrotate?) %s\n", log);
+	    continue;
+	 }
+	 time_t mtime = buf.st_mtime;
+	 g_debug_update ("mtime from %s: %i (%s)\n", log, mtime, ctime(&mtime));
+	 time_t bctime = buf.st_ctime;
+	 g_debug_update ("ctime from %s: %i (%s)\n", log, bctime, ctime(&bctime));
+	 newest_log_stamp = MAX(MAX(mtime, bctime), newest_log_stamp);
+	 g_debug_update ("last_launch from %s: %i (%s)\n", log, newest_log_stamp, ctime(&newest_log_stamp));
+      }
+      globfree(&pglob);
+   }
+   return newest_log_stamp;
+}
+
+// check if the auto launch interval is over and its 
+// time to launch again and if the dpkg lock is currently 
+// not taken
+static gboolean 
+auto_launch_now (UpdateTrayAppletPrivate *priv)
+{
+   int interval_days = 0;
+   time_t last_launch = 0;
+
+   if (dpkg_lock_is_taken())
+      return FALSE;
+
+#if 0 // disabled because we need more cleverness here in order to cover
+      // the case where people always work on battery (and charge overnight)
+   if (running_on_battery())
+      return FALSE;
+#endif
+
+   // when checking for regular updates honor the 
+   // regular_auto_launch_interval
+   interval_days = gconf_client_get_int(priv->gconf,
+					GCONF_KEY_AUTO_LAUNCH_INTERVAL, 
+					NULL);
+   g_debug_update ("interval_days from gconf: %i\n", interval_days);
+
+   if (interval_days <= 0) 
+      return TRUE;
+
+   // check last launch time 
+   last_launch = gconf_client_get_int(priv->gconf,
+				      GCONF_KEY_LAST_LAUNCH,
+				      NULL);
+   g_debug_update ("last_launch from gconf: %i (%s)\n", last_launch, ctime(&last_launch));
+
+   time_t now = time(NULL);
+   if (auto_launch_security_now(priv, now, last_launch))
+      return TRUE;
+
+   time_t log_files = newest_log_file_timestamp();
+   last_launch = MAX(log_files, last_launch);
+
+   g_debug_update("time now %i (%s), delta: %i", now, ctime(&now), now-last_launch);
+   if ((last_launch + (24*60*60*interval_days)) < now) {
+      g_debug_update ("need to auto launch");
+      return TRUE;
+   }
+
+   return FALSE;
+}
+
 gboolean
 update_check (TrayApplet *ta)
 {
    if(ta == NULL)
       return FALSE;
 
-   //g_print("update_check()\n");
+   g_debug_update ("update_check()\n");
    UpdateTrayAppletPrivate *priv = (UpdateTrayAppletPrivate*)ta->user_data;
 
-   int num_upgrades = 0;
-   int num_security = 0;
+   priv->num_upgrades = 0;
+   priv->num_security = 0;
    int exit_status = 0;
    GError *error;
 
@@ -341,20 +598,22 @@ update_check (TrayApplet *ta)
 	 g_strfreev(ret_array);
 	 return TRUE;
       }
-      num_upgrades = atoi(ret_array[0]);
-      num_security = atoi(ret_array[1]);
+      priv->num_upgrades = atoi(ret_array[0]);
+      priv->num_security = atoi(ret_array[1]);
       g_free(ret);
       g_strfreev(ret_array);
    } else
       g_warning("Error launching %s", UPGRADE_CHECKER);
    
-   if (num_upgrades == 0) {
+   g_debug_update("%s returned %i (security: %i)", UPGRADE_CHECKER, priv->num_upgrades, priv->num_security);
+
+   if (priv->num_upgrades == 0) {
 
       // check if the periodic file is very old
       struct stat buf;
       if ((stat("/var/lib/apt/periodic/update-success-stamp", &buf) == 0) &&
 	  (time(NULL) - buf.st_mtime > OUTDATED_NAG_AGE) ) 
-	 g_timeout_add_seconds(OUTDATED_NAG_WAIT, outdated_nag, ta);
+	 g_timeout_add_seconds(OUTDATED_NAG_WAIT, (GSourceFunc)outdated_nag, ta);
 
       gtk_status_icon_set_visible (ta->tray_icon, FALSE);
       
@@ -365,14 +624,28 @@ update_check (TrayApplet *ta)
       return TRUE;
    } 
 
-   if (num_security == 0)
+   if (priv->num_security == 0)
       ta->name = "software-update-available";
    else
       ta->name = "software-update-urgent";
    gtk_status_icon_set_from_icon_name(ta->tray_icon, ta->name);
 
    // update the tooltip
-   update_trayicon_update_tooltip(ta, num_upgrades);
+   update_trayicon_update_tooltip(ta, priv->num_upgrades);
+
+   // check if the user wants to see the icon or launch
+   // the default action
+   if(gconf_client_get_bool(priv->gconf,
+			    GCONF_KEY_AUTO_LAUNCH, NULL)) 
+   {
+      gtk_status_icon_set_visible(ta->tray_icon, FALSE);
+      if (auto_launch_now(priv)) 
+      {
+	 g_spawn_command_line_async("nice ionice -c3 update-manager "
+				    "--no-focus-on-map", NULL);
+      }
+      return TRUE;
+   }
 
    // if we are already visible, skip the rest
    if(gtk_status_icon_get_visible (ta->tray_icon))
@@ -380,7 +653,7 @@ update_check (TrayApplet *ta)
 
    // show the icon
    gtk_status_icon_set_visible(ta->tray_icon, TRUE);
-
+   
    // the user does not no notification messages
    if(gconf_client_get_bool(priv->gconf,
 			    GCONF_KEY_NO_UPDATE_NOTIFICATIONS, NULL))
@@ -388,8 +661,8 @@ update_check (TrayApplet *ta)
 
    // show the notification with some delay. otherwise on a login
    // the origin of the window is 0,0 and that looks ugly
-   g_timeout_add(5000, show_notification, ta);
-   
+   g_timeout_add_seconds(5, show_notification, ta);
+
    return TRUE;
 }
 
